@@ -282,14 +282,14 @@ overlayUtils.addLayer = function(layerName, tileSource, i, visible) {
             }
             interfaceUtils.alert("Impossible to load file.")
             showModal = false;
-        }
+        } 
     });
 }
 
 
 /** 
  * @param {Number} item Index of an OSD tile source
- * Set the opacity of a tile source */
+ * @summary Set the opacity of a tile source */
 overlayUtils.setItemOpacity = function(item) {
     opacity = overlayUtils._layerOpacities[item];
 
@@ -339,7 +339,7 @@ overlayUtils.waitFullyLoaded = function () {
 /** 
  * @param {String} layerName name of an existing d3 node
  * @param {Number} opacity desired opacity
- * Set the opacity of a tile source */
+ * @summary Set the opacity of a tile source */
 overlayUtils.setLayerOpacity= function(layerName,opacity){
     if(layerName in overlayUtils._d3nodes){
         var layer = overlayUtils._d3nodes[layerName];
@@ -351,7 +351,7 @@ overlayUtils.setLayerOpacity= function(layerName,opacity){
 
 /**
  * @param {String} colortype A string from [hex,hsl,rgb]
- * Get a random color in the desired format
+ * @summary Get a random color in the desired format
  */
 overlayUtils.randomColor = function (colortype) {
     if (!colortype) {
@@ -437,15 +437,18 @@ overlayUtils.saveSVG=function(){
  * Save the current canvas as a PNG image
  */
 overlayUtils.savePNG=function() {
-    interfaceUtils.prompt("Resolution for export (1 = screen resolution):<br/><small><i>Max output size: 4096x4096 pixels</i></small><br/><br/><small>High resolution can take time to load!</small>","5","Capture viewport","number")
+    interfaceUtils.prompt("Resolution for export (1 = screen resolution):<br/>High resolution can take time to load!</small>","4","Capture viewport","number")
     .then((resolution) => {
         var bounds = tmapp.ISS_viewer.viewport.getBounds();
         var loading=interfaceUtils.loadingModal();
         tmapp.ISS_viewer.world.getItemAt(0).immediateRender = true
+        var strokeWidth = regionUtils._polygonStrokeWidth
+        regionUtils._polygonStrokeWidth *= resolution
         overlayUtils.waitFullyLoaded().then(() => {
             overlayUtils.getCanvasPNG(resolution)
             .then (() => {
                 // We go back to original size:
+                regionUtils._polygonStrokeWidth = strokeWidth;
                 tmapp.ISS_viewer.world.getItemAt(0).immediateRender = false
                 tmapp.ISS_viewer.viewport.fitBounds(bounds, true);
                 setTimeout(()=>{$(loading).modal("hide");}, 300);
@@ -507,7 +510,7 @@ overlayUtils.savePNG=function() {
             var canvas = document.createElement("canvas");
             var ctx = canvas.getContext("2d");
             var ctx_osd = document.querySelector(".openseadragon-canvas canvas").getContext("2d");
-            var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl");
+            var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl2");
             canvas.width = tiling * Math.min(ctx_osd.canvas.width, ctx_webgl.canvas.width);
             canvas.height = tiling * Math.min(ctx_osd.canvas.height, ctx_webgl.canvas.height);
             var bounds = tmapp.ISS_viewer.viewport.getBounds();
@@ -543,7 +546,7 @@ overlayUtils.savePNG=function() {
         // Create an empty canvas element
         var canvas = document.createElement("canvas");
         var ctx_osd = document.querySelector(".openseadragon-canvas canvas").getContext("2d");
-        var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl");
+        var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl2");
         canvas.width = Math.min(ctx_osd.canvas.width, ctx_webgl.canvas.width);
         canvas.height = Math.min(ctx_osd.canvas.height, ctx_webgl.canvas.height);
 
