@@ -3,12 +3,17 @@
 * @author Fredrik Nysjo
 * @see {@link glUtils}
 */
+
+/**
+ * @namespace glUtils
+ * @property {Boolean} _initialized True when glUtils has been initialized
+ */
 glUtils = {
     _initialized: false,
     _imageSize: [1, 1],
     _viewportRect: [0, 0, 1, 1],
-    _options: {antialias: false, premultipliedAlpha: false, preserveDrawingBuffer: true},
-
+    _options: {antialias: false, premultipliedAlpha: true, preserveDrawingBuffer: true},
+    _markershapes: "misc/markershapes.png",
     // WebGL resources
     _programs: {},
     _buffers: {},
@@ -22,6 +27,7 @@ glUtils = {
     _markerScaleFactor: {},      // {uid: float}
     _markerScalarPropertyName: {},  // {uid: string, ...}
     _markerOpacity: {},          // {uid: alpha, ...}
+    _markerOutline: {},          // {uid: boolean, ...}
     _useColorFromMarker: {},     // {uid: boolean, ...}
     _useColorFromColormap: {},   // {uid: boolean, ...}
     _useScaleFromMarker: {},     // {uid: boolean, ...}
@@ -42,6 +48,8 @@ glUtils = {
     _pickedMarker: [-1, -1],
     _showColorbar: true,
     _showMarkerInfo: true,
+    _resolutionScale: 1.0,        // If this is set to below 1.0, the WebGL output will be upscaled
+    _resolutionScaleActual: 1.0,  // Automatic scaling factor computed from glUtils._resolutionScale
     _piechartPalette: ["#fff100", "#ff8c00", "#e81123", "#ec008c", "#68217a", "#00188f", "#00bcf2", "#00b294", "#009e49", "#bad80a"]
 }
 
@@ -139,6 +147,7 @@ glUtils._markersVS = `
 glUtils._markersFS = `
     precision mediump float;
 
+    uniform bool u_markerOutline;
     uniform bool u_usePiechartFromMarker;
     uniform bool u_alphaPass;
     uniform sampler2D u_shapeAtlas;
@@ -177,7 +186,13 @@ glUtils._markersFS = `
         vec2 uv = (gl_PointCoord.xy - 0.5) * UV_SCALE + 0.5;
         uv = (uv + v_shapeOrigin) * (1.0 / SHAPE_GRID_SIZE);
 
+        // Sample shape texture in which the blue channel encodes alpha and the
+        // red and green channels encode grayscale for marker shape with and
+        // without outline, respectively
         vec4 shapeColor = texture2D(u_shapeAtlas, uv, -0.5);
+        shapeColor = u_markerOutline ? shapeColor.rrrb : shapeColor.gggb;
+
+        // This bias avoids minified markers with outline becoming too dark
         float shapeColorBias = max(0.0, 1.0 - v_shapeSize * 0.2);
         shapeColor.rgb = clamp(shapeColor.rgb + shapeColorBias, 0.0, 1.0);
 
@@ -267,12 +282,13 @@ glUtils._pickingVS = `
             if (abs(uv.x - 0.5) > 0.5 || abs(uv.y - 0.5) > 0.5) DISCARD_VERTEX;
 
             // Do fine-grained inside/outside test by sampling the shape texture
+            // with alpha encoded in the blue channel
             vec2 shapeOrigin = vec2(0.0);
             shapeOrigin.x = mod((shapeID + 0.00001) * 255.0 - 1.0, SHAPE_GRID_SIZE);
             shapeOrigin.y = floor(((shapeID + 0.00001) * 255.0 - 1.0) / SHAPE_GRID_SIZE);
             uv = (uv - 0.5) * UV_SCALE + 0.5;
             uv = (uv + shapeOrigin) * (1.0 / SHAPE_GRID_SIZE);
-            if (texture2D(u_shapeAtlas, uv).r < 0.5) DISCARD_VERTEX;
+            if (texture2D(u_shapeAtlas, uv).b < 0.5) DISCARD_VERTEX;
 
             // Also do a quick alpha-test to avoid picking non-visible markers
             if (a_opacity * u_markerOpacity <= 0.0) DISCARD_VERTEX
@@ -356,7 +372,10 @@ glUtils._createPiechartAngles = function(sectors) {
 }
 
 
-// Load markers loaded from CSV file into vertex buffer
+/**
+ * @summary Create WebGL resources and other objects for drawing marker dataset.
+ * @param {String | Number} uid Identifier referencing the marker dataset in dataUtils.
+ */
 glUtils.loadMarkers = function(uid) {
     if (!glUtils._initialized) return;
     const canvas = document.getElementById("gl_canvas");
@@ -407,11 +426,13 @@ glUtils.loadMarkers = function(uid) {
     const useShapeFromMarker = dataUtils.data[uid]["_shape_col"] != null;
     const numShapes = Object.keys(markerUtils._symbolStrings).length;
     let shapeIndex = 0;
-    
+
     const opacityPropertyName = dataUtils.data[uid]["_opacity_col"];
     const useOpacityFromMarker = dataUtils.data[uid]["_opacity_col"] != null;
     const markerOpacityFactor = dataUtils.data[uid]["_opacity"];
 
+    const markerOutline = !dataUtils.data[uid]["_no_outline"];
+
     // Additional info about the vertex format
     const NUM_COMPONENTS_PER_MARKER = 8;
     const NUM_BYTES_PER_MARKER = NUM_COMPONENTS_PER_MARKER * 4;
@@ -533,17 +554,18 @@ glUtils.loadMarkers = function(uid) {
 
     // Update marker info and LUT + colormap textures
     glUtils._numPoints[uid] = numPoints * numSectors;
-    glUtils._markerOpacity[uid] = markerOpacityFactor;
     glUtils._markerScalarRange[uid] = scalarRange;
     glUtils._markerScalarPropertyName[uid] = scalarPropertyName;
     glUtils._markerScaleFactor[uid] = markerScaleFactor;
-    glUtils._colorscaleName[uid] = colorscaleName;
+    glUtils._markerOpacity[uid] = markerOpacityFactor;
+    glUtils._markerOutline[uid] = markerOutline;
     glUtils._useColorFromMarker[uid] = useColorFromMarker;
     glUtils._useColorFromColormap[uid] = useColorFromColormap;
     glUtils._useScaleFromMarker[uid] = useScaleFromMarker;
     glUtils._useOpacityFromMarker[uid] = useOpacityFromMarker;
     glUtils._usePiechartFromMarker[uid] = usePiechartFromMarker;
     glUtils._useShapeFromMarker[uid] = useShapeFromMarker;
+    glUtils._colorscaleName[uid] = colorscaleName;
     if (useColorFromColormap) {
         glUtils._updateColorScaleTexture(gl, uid, glUtils._textures[uid + "_colorscale"]);
     }
@@ -553,6 +575,10 @@ glUtils.loadMarkers = function(uid) {
 }
 
 
+/**
+ * @summary Delete WebGL resources and other objects created for drawing marker dataset.
+ * @param {String | Number} uid Identifier referencing the marker dataset in dataUtils.
+ */
 glUtils.deleteMarkers = function(uid) {
     if (!glUtils._initialized) return;
     const canvas = document.getElementById("gl_canvas");
@@ -564,7 +590,9 @@ glUtils.deleteMarkers = function(uid) {
     delete glUtils._numPoints[uid];
     delete glUtils._markerScaleFactor[uid];
     delete glUtils._markerScalarRange[uid];
+    delete glUtils._markerScalarPropertyName[uid];
     delete glUtils._markerOpacity[uid];
+    delete glUtils._markerOutline[uid];
     delete glUtils._useColorFromMarker[uid];
     delete glUtils._useColorFromColormap[uid];
     delete glUtils._useScaleFromMarker[uid];
@@ -666,6 +694,10 @@ glUtils._updateColorLUTTexture = function(gl, uid, texture) {
 }
 
 
+/**
+ * @summary Update the color scale LUTs for all marker datasets.
+ * This function is a callback and should not normally be called directly.
+ */
 glUtils.updateColorLUTTextures = function() {
     const canvas = document.getElementById("gl_canvas");
     const gl = canvas.getContext("webgl", glUtils._options);
@@ -836,7 +868,7 @@ glUtils._createMarkerWebGLCanvas = function() {
     const canvas = document.createElement("canvas");
     canvas.id = "gl_canvas";
     canvas.width = "1"; canvas.height = "1";
-    canvas.style = "position:relative; pointer-events:none; z-index: 12;";
+    canvas.style = "position:relative; pointer-events:none; z-index: 12; width: 100%; height: 100%";
     return canvas;
 }
 
@@ -860,7 +892,7 @@ glUtils._loadTextureFromImageURL = function(gl, src) {
 }
 
 
-glUtils.drawColorPass = function(gl, viewportTransform, markerScaleAdjusted) {
+glUtils._drawColorPass = function(gl, viewportTransform, markerScaleAdjusted) {
     // Set up render pipeline
     const program = glUtils._programs["markers"];
     gl.useProgram(program);
@@ -902,6 +934,7 @@ glUtils.drawColorPass = function(gl, viewportTransform, markerScaleAdjusted) {
         gl.uniform1f(gl.getUniformLocation(program, "u_globalMarkerScale"), glUtils._globalMarkerScale * glUtils._markerScaleFactor[uid]);
         gl.uniform2fv(gl.getUniformLocation(program, "u_markerScalarRange"), glUtils._markerScalarRange[uid]);
         gl.uniform1f(gl.getUniformLocation(program, "u_markerOpacity"), glUtils._markerOpacity[uid]);
+        gl.uniform1i(gl.getUniformLocation(program, "u_markerOutline"), glUtils._markerOutline[uid]);
         gl.uniform1i(gl.getUniformLocation(program, "u_useColorFromMarker"), glUtils._useColorFromMarker[uid]);
         gl.uniform1i(gl.getUniformLocation(program, "u_useColorFromColormap"), glUtils._useColorFromColormap[uid]);
         gl.uniform1i(gl.getUniformLocation(program, "u_usePiechartFromMarker"), glUtils._usePiechartFromMarker[uid]);
@@ -938,7 +971,7 @@ glUtils.drawColorPass = function(gl, viewportTransform, markerScaleAdjusted) {
 }
 
 
-glUtils.drawPickingPass = function(gl, viewportTransform, markerScaleAdjusted) {
+glUtils._drawPickingPass = function(gl, viewportTransform, markerScaleAdjusted) {
     // Set up render pipeline
     const program = glUtils._programs["picking"];
     gl.useProgram(program);
@@ -1002,6 +1035,12 @@ glUtils.drawPickingPass = function(gl, viewportTransform, markerScaleAdjusted) {
 }
 
 
+/**
+ * @summary Do rendering to the WebGL canvas.
+ * Calling this function will force an update of the rendering of markers and
+ * the data used for picking (i.e. for marker selection). Only marker datasets
+ * for which glUtils.loadMarkers() have been called will be rendered.
+ */
 glUtils.draw = function() {
     const canvas = document.getElementById("gl_canvas");
     const gl = canvas.getContext("webgl", glUtils._options);
@@ -1021,23 +1060,31 @@ glUtils.draw = function() {
     let markerScaleAdjusted = glUtils._markerScale;
     if (glUtils._useMarkerScaleFix) markerScaleAdjusted *= (gl.canvas.height / 900.0);
 
-    gl.clearColor(0.5, 0.5, 0.5, 0.0);
+    gl.clearColor(0.0, 0.0, 0.0, 0.0);
     gl.clear(gl.COLOR_BUFFER_BIT);
     gl.viewport(0, 0, gl.canvas.width, gl.canvas.height);
 
     if (glUtils._pickingEnabled) {
-        glUtils.drawPickingPass(gl, viewportTransform, markerScaleAdjusted);
+        glUtils._drawPickingPass(gl, viewportTransform, markerScaleAdjusted);
         glUtils._pickingEnabled = false;  // Clear flag until next click event
     }
 
-    glUtils.drawColorPass(gl, viewportTransform, markerScaleAdjusted);
+    glUtils._drawColorPass(gl, viewportTransform, markerScaleAdjusted);
 }
 
 
+/**
+ * @summary Do GPU-based picking for marker selection.
+ * This function is a callback and should not normally be called directly. The
+ * function will automatically call glUtils.draw() to update the rendering and
+ * the picking.
+ * @param {Object} event An object with click events from the canvas
+ */
 glUtils.pick = function(event) {
     if (event.quick) {
         glUtils._pickingEnabled = true;
-        glUtils._pickingLocation = [event.position.x, event.position.y];
+        glUtils._pickingLocation = [event.position.x * glUtils._resolutionScaleActual,
+                                    event.position.y * glUtils._resolutionScaleActual];
         glUtils.draw();  // This will update the value of glUtils._pickedMarker
 
         const pickedMarker = glUtils._pickedMarker;
@@ -1057,7 +1104,7 @@ glUtils.pick = function(event) {
             div.id = "ISS_marker_info";
             div.width = "1px"; div.height = "1px";
             if (glUtils._usePiechartFromMarker[uid]) {
-                div.innerHTML = markerUtils.makePiechartTable(markerData, markerIndex, piechartPropertyName);
+                div.innerHTML = markerUtils.makePiechartTable(uid, markerIndex, piechartPropertyName);
             } else {
                 div.innerHTML = markerUtils.getMarkerTooltip(uid, markerIndex);
                 console.log("Marker clicked:",tabName, groupName, "index:", markerIndex);
@@ -1084,23 +1131,48 @@ glUtils.pick = function(event) {
 }
 
 
+/**
+ * @summary Callback for resizing the WebGL canvas.
+ * Calling this function will force an update of the width and height of the
+ * WebGL canvas, but will not automatically call glUtils.draw() to update the
+ * rendering.
+ */
 glUtils.resize = function() {
     const canvas = document.getElementById("gl_canvas");
     const gl = canvas.getContext("webgl", glUtils._options);
 
     const op = tmapp["object_prefix"];
-    const newSize = tmapp[op + "_viewer"].viewport.containerSize;
-    gl.canvas.width = newSize.x;
-    gl.canvas.height = newSize.y;
+    const width = tmapp[op + "_viewer"].viewport.containerSize.x;
+    const height = tmapp[op + "_viewer"].viewport.containerSize.y;
+
+    glUtils._resolutionScaleActual = glUtils._resolutionScale * window.devicePixelRatio;
+    if (Math.max(width, height) * glUtils._resolutionScale >= 4096.0) {
+        // A too large WebGL canvas can lead to misalignment between the WebGL
+        // markers and the OSD image layers, so here the resolution scaling
+        // factor is adjusted to restrict the canvas size to a safe value
+        glUtils._resolutionScaleActual *= 4096.0 / (Math.max(width, height) * glUtils._resolutionScale);
+    }
+
+    gl.canvas.width = width * glUtils._resolutionScaleActual;
+    gl.canvas.height = height * glUtils._resolutionScaleActual;
 }
 
 
+/**
+ * @summary Callback for resizing the WebGL canvas.
+ * Works like glUtils.resize(), but will also automatically call glUtils.draw()
+ * to update the rendering.
+ */
 glUtils.resizeAndDraw = function() {
     glUtils.resize();
     glUtils.draw();
 }
 
 
+/**
+ * @summary Callback for updating marker scale when changing the global marker size GUI slider.
+ * This function is a callback and should not normally be called directly.
+ */
 glUtils.updateMarkerScale = function() {
     const globalMarkerSize = Number(document.getElementById("ISS_globalmarkersize_text").value);
     // Clamp the scale factor to avoid giant markers and slow rendering if the
@@ -1109,7 +1181,12 @@ glUtils.updateMarkerScale = function() {
 }
 
 
-glUtils._restoreLostContext = function(event) {
+/**
+ * @summary Callback for restoring WebGL resources after WebGL context is lost
+ * This function is a callback and should not normally be called directly. Loss
+ * of context can happen when for example the computer goes into sleep mode.
+ */
+glUtils.restoreLostContext = function(event) {
     console.log("Restoring WebGL objects after context loss");
     let canvas = document.getElementById("gl_canvas");
     const gl = canvas.getContext("webgl", glUtils._options);
@@ -1117,7 +1194,7 @@ glUtils._restoreLostContext = function(event) {
     // Restore shared WebGL objects
     glUtils._programs["markers"] = glUtils._loadShaderProgram(gl, glUtils._markersVS, glUtils._markersFS);
     glUtils._programs["picking"] = glUtils._loadShaderProgram(gl, glUtils._pickingVS, glUtils._pickingFS);
-    glUtils._textures["shapeAtlas"] = glUtils._loadTextureFromImageURL(gl, "misc/markershapes.png");
+    glUtils._textures["shapeAtlas"] = glUtils._loadTextureFromImageURL(gl, glUtils._markershapes);
 
     // Restore per-markers WebGL objects
     for (let [uid, numPoints] of Object.entries(glUtils._numPoints)) {
@@ -1131,13 +1208,19 @@ glUtils._restoreLostContext = function(event) {
 }
 
 
+/**
+ * @summary Do initialization of the WebGL canvas.
+ * This will also load WebGL resources like shaders and textures, as well as set
+ * up events for interaction with other parts of TissUUmaps such as the
+ * OpenSeaDragon (OSD) canvas.
+ */
 glUtils.init = function() {
     if (glUtils._initialized) return;
 
     let canvas = document.getElementById("gl_canvas");
     if (!canvas) canvas = this._createMarkerWebGLCanvas();
     canvas.addEventListener("webglcontextlost", function(e) { e.preventDefault(); }, false);
-    canvas.addEventListener("webglcontextrestored", glUtils._restoreLostContext, false);
+    canvas.addEventListener("webglcontextrestored", glUtils.restoreLostContext, false);
     const gl = canvas.getContext("webgl", glUtils._options);
 
     // Place marker canvas under the OSD canvas. Doing this also enables proper
@@ -1147,7 +1230,7 @@ glUtils.init = function() {
 
     this._programs["markers"] = this._loadShaderProgram(gl, this._markersVS, this._markersFS);
     this._programs["picking"] = this._loadShaderProgram(gl, this._pickingVS, this._pickingFS);
-    this._textures["shapeAtlas"] = this._loadTextureFromImageURL(gl, "misc/markershapes.png");
+    this._textures["shapeAtlas"] = this._loadTextureFromImageURL(gl, glUtils._markershapes);
 
     this._createColorbarCanvas();  // The colorbar is drawn separately in a 2D-canvas
 
