@@ -187,8 +187,6 @@ overlayUtils.addLayerSettings = function(layerName, tileSource, layerIndex, chec
         $(elt).bind(mousewheelevt, moveSlider);
         function moveSlider(e){
             var zoomLevel = parseInt($(channelRange).val()); 
-            console.log("Mouse wheel!", zoomLevel, zoomLevel+1, zoomLevel-1)
-            console.log(e.originalEvent.wheelDelta);
             // detect positive or negative scrolling
             if ( e.originalEvent.wheelDelta < 0 ) {
                 //scroll down
@@ -234,7 +232,6 @@ overlayUtils.addLayerFromSelect = function() {
         name: layerName,
         tileSource: tileSource
     });
-    console.log("tileSource", tileSource);
     i = tmapp.layers.length - 1;
     overlayUtils.addLayer(layerName, tileSource, i);
     overlayUtils.addAllLayersSettings();
@@ -243,12 +240,29 @@ overlayUtils.addLayerFromSelect = function() {
 /**
  * This method is used to add a layer */
 overlayUtils.addLayer = function(layerName, tileSource, i, visible) {
+    const queryString = window.location.search;
+    const urlParams = new URLSearchParams(queryString);
+    const path = urlParams.get('path')
+    if (path != null) {
+        tileSource = path + "/" + tileSource
+    }
     var op = tmapp["object_prefix"];
     var vname = op + "_viewer";
     var opacity = 1.0;
     if (i >= 0 && !visible) {
         opacity = 0.0;
     }
+    var showModal = true;
+    var loadingModal = null;
+    setTimeout(function(){
+        if (showModal)
+            loadingModal = interfaceUtils.loadingModal("Converting image, please wait...");
+    },800);
+    if (tmapp["ISS_viewer"].world.getItemCount() != 0) {
+        if (tmapp["ISS_viewer"].world.getItemAt(0).source.getTileUrl(0,0,0) == null) {
+            tmapp["ISS_viewer"].close();
+        }
+    }
     tmapp[vname].addTiledImage({
         index: i + 1,
         tileSource: tmapp._url_suffix + tileSource,
@@ -257,6 +271,17 @@ overlayUtils.addLayer = function(layerName, tileSource, i, visible) {
             layer0X = tmapp[op + "_viewer"].world.getItemAt(0).getContentSize().x;
             layerNX = tmapp[op + "_viewer"].world.getItemAt(tmapp[op + "_viewer"].world.getItemCount()-1).getContentSize().x;
             tmapp[op + "_viewer"].world.getItemAt(tmapp[op + "_viewer"].world.getItemCount()-1).setWidth(layerNX/layer0X);
+            if (loadingModal) {
+                setTimeout(function(){$(loadingModal).modal("hide");}, 500);
+            }
+            showModal = false;
+        },
+        error: function(i) {
+            if (loadingModal) {
+                setTimeout(function(){$(loadingModal).modal("hide");}, 500);
+            }
+            interfaceUtils.alert("Impossible to load file.")
+            showModal = false;
         } 
     });
 }
@@ -264,7 +289,7 @@ overlayUtils.addLayer = function(layerName, tileSource, i, visible) {
 
 /** 
  * @param {Number} item Index of an OSD tile source
- * Set the opacity of a tile source */
+ * @summary Set the opacity of a tile source */
 overlayUtils.setItemOpacity = function(item) {
     opacity = overlayUtils._layerOpacities[item];
 
@@ -291,10 +316,30 @@ overlayUtils.areAllFullyLoaded = function () {
     return true;
   }
 
+overlayUtils.waitFullyLoaded = function () {
+    function sleep (time) {
+        return new Promise((resolve) => setTimeout(resolve, time));
+    }
+    return new Promise((resolve, reject) => {
+        sleep(200).then (()=>{
+            if (overlayUtils.areAllFullyLoaded()) {
+                resolve();
+                return;
+            }
+            else {
+                overlayUtils.waitFullyLoaded().then(()=>{
+                    resolve();
+                    return;
+                });
+            }    
+        });
+    });
+}
+
 /** 
  * @param {String} layerName name of an existing d3 node
  * @param {Number} opacity desired opacity
- * Set the opacity of a tile source */
+ * @summary Set the opacity of a tile source */
 overlayUtils.setLayerOpacity= function(layerName,opacity){
     if(layerName in overlayUtils._d3nodes){
         var layer = overlayUtils._d3nodes[layerName];
@@ -306,12 +351,13 @@ overlayUtils.setLayerOpacity= function(layerName,opacity){
 
 /**
  * @param {String} colortype A string from [hex,hsl,rgb]
- * Get a random color in the desired format
+ * @summary Get a random color in the desired format
  */
-overlayUtils.randomColor = function (colortype) {
+overlayUtils.randomColor = function (colortype, seed) {
     if (!colortype) {
         colortype = "hex";
     }
+    return randomColor({luminosity: 'bright', format: colortype, seed:seed});
     //I need random colors that are far away from the palette in the image
     //in this case Hematoxilyn and DAB so far away from brown and light blue
     //and avoid light colors because of the white  background 
@@ -392,40 +438,136 @@ overlayUtils.saveSVG=function(){
  * Save the current canvas as a PNG image
  */
 overlayUtils.savePNG=function() {
-    // Create an empty canvas element
-    $("#loadingModal").show();
-    var canvas = document.createElement("canvas");
-    var ctx_osd = document.querySelector(".openseadragon-canvas canvas").getContext("2d");
-    var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl");
-    canvas.width = ctx_osd.canvas.width;
-    canvas.height = ctx_osd.canvas.height;
-    
-    // Copy the image contents to the canvas
-    var ctx = canvas.getContext("2d");
-    
-    ctx.drawImage(ctx_osd.canvas, 0, 0);
-    ctx.drawImage(ctx_webgl.canvas, 0, 0);
-    console.log(canvas.width,canvas.height);
-    var dataURL = canvas.toDataURL("image/png");
-    
-    var svgString = new XMLSerializer().serializeToString(document.querySelector('.openseadragon-canvas svg'));
+    interfaceUtils.prompt("Resolution for export (1 = screen resolution):<br/>High resolution can take time to load!</small>","4","Capture viewport","number")
+    .then((resolution) => {
+        var bounds = tmapp.ISS_viewer.viewport.getBounds();
+        var loading=interfaceUtils.loadingModal();
+        tmapp.ISS_viewer.world.getItemAt(0).immediateRender = true
+        var strokeWidth = regionUtils._polygonStrokeWidth
+        regionUtils._polygonStrokeWidth *= resolution
+        overlayUtils.waitFullyLoaded().then(() => {
+            overlayUtils.getCanvasPNG(resolution)
+            .then (() => {
+                // We go back to original size:
+                regionUtils._polygonStrokeWidth = strokeWidth;
+                tmapp.ISS_viewer.world.getItemAt(0).immediateRender = false
+                tmapp.ISS_viewer.viewport.fitBounds(bounds, true);
+                setTimeout(()=>{$(loading).modal("hide");}, 300);
+                
+                document.getElementById("ISS_viewer").style.setProperty("visibility", "unset");
+            })
+        });
+    })
+}
 
-    var DOMURL = self.URL || self.webkitURL || self;
-    var img = new Image();
-    var svg = new Blob([svgString], {type: "image/svg+xml;charset=utf-8"});
-    var url = DOMURL.createObjectURL(svg);
-    img.onload = function() {
-        ctx.drawImage(img, 0, 0);
-        var png = canvas.toDataURL("image/png");
-        console.log(png.replace(/^data:image\/(png|jpg);base64,/, ""));
-    
-        var a = document.createElement("a"); //Create <a>
-        a.href = png; //Image Base64 Goes here
-        a.download = "TissUUmaps_capture.png"; //File name Here
-        a.click(); //Downloaded file
-        $("#loadingModal").hide();
-        DOMURL.revokeObjectURL(png);
-    };
-    img.src = url;
-    console.log(url);
+/**
+ * Get the current canvas as a PNG image
+ */
+ overlayUtils.getCanvasPNG=function(tiling) {
+    tiling = tiling ? Math.ceil(tiling) : 1;
+    function sleep (time) {
+        return new Promise((resolve) => setTimeout(resolve, time));
+    }
+    function getCanvasCtx_aux (index, size, ctx, bounds) {
+        return new Promise((resolve, reject) => {
+            if (index == size*size) {
+                resolve(ctx);
+                return;
+            }
+            var index_x = index % size;
+            var index_y = Math.floor(index/size);
+            var newBounds = new OpenSeadragon.Rect(
+                bounds.x+index_x*bounds.width/size,
+                bounds.y+index_y*bounds.height/size,
+                bounds.width/size,
+                bounds.height/size,
+                0
+            );
+            console.log(bounds, newBounds);
+            tmapp.ISS_viewer.viewport.fitBounds(newBounds, true);
+            overlayUtils.waitFullyLoaded().then(() => {
+                overlayUtils.getCanvasCtx().then ((ctx_offset) => {
+                    ctx.drawImage(
+                        ctx_offset.canvas, 
+                        ctx_offset.canvas.width * index_x, 
+                        ctx_offset.canvas.height * index_y, 
+                        ctx_offset.canvas.width,
+                        ctx_offset.canvas.height
+                    );
+                    getCanvasCtx_aux(index+1, size, ctx, bounds).then(
+                        (ctx)=>{
+                            resolve(ctx);
+                            return;
+                        }
+                        
+                    )
+                });
+            });
+        });
+    }
+
+    return new Promise((resolve, reject) => {
+        if (tiling > 1) {
+            var canvas = document.createElement("canvas");
+            var ctx = canvas.getContext("2d");
+            var ctx_osd = document.querySelector(".openseadragon-canvas canvas").getContext("2d");
+            var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl");
+            canvas.width = tiling * Math.min(ctx_osd.canvas.width, ctx_webgl.canvas.width);
+            canvas.height = tiling * Math.min(ctx_osd.canvas.height, ctx_webgl.canvas.height);
+            var bounds = tmapp.ISS_viewer.viewport.getBounds();
+            getCanvasCtx_aux(0, tiling, ctx, bounds).then((ctx_tiling) => {
+                var png = ctx_tiling.canvas.toDataURL("image/png");
+                
+                var a = document.createElement("a"); //Create <a>
+                a.href = png; //Image Base64 Goes here
+                a.download = "TissUUmaps_capture.png"; //File name Here
+                a.click(); //Downloaded file
+                resolve(png);
+            })
+        }
+        else {
+            overlayUtils.getCanvasCtx().then((ctx)  =>{
+                var png = ctx.canvas.toDataURL("image/png");
+                
+                var a = document.createElement("a"); //Create <a>
+                a.href = png; //Image Base64 Goes here
+                a.download = "TissUUmaps_capture.png"; //File name Here
+                a.click(); //Downloaded file
+                resolve(png);
+            })
+        }
+    })
+}
+
+/**
+ * Get the current canvas as a 2d context
+ */
+ overlayUtils.getCanvasCtx=function() {
+    return new Promise((resolve, reject) => {
+        // Create an empty canvas element
+        var canvas = document.createElement("canvas");
+        var ctx_osd = document.querySelector(".openseadragon-canvas canvas").getContext("2d");
+        var ctx_webgl = document.querySelector("#gl_canvas").getContext("webgl");
+        canvas.width = Math.min(ctx_osd.canvas.width, ctx_webgl.canvas.width);
+        canvas.height = Math.min(ctx_osd.canvas.height, ctx_webgl.canvas.height);
+
+        // Copy the image contents to the canvas
+        var ctx = canvas.getContext("2d");
+        
+        ctx.drawImage(ctx_osd.canvas, 0, 0, canvas.width, canvas.height);
+        ctx.drawImage(ctx_webgl.canvas, 0, 0, canvas.width, canvas.height);
+        
+        var svgString = new XMLSerializer().serializeToString(document.querySelector('.openseadragon-canvas svg'));
+
+        var DOMURL = self.URL || self.webkitURL || self;
+        var img = new Image();
+        var svg = new Blob([svgString], {type: "image/svg+xml;charset=utf-8"});
+        var url = DOMURL.createObjectURL(svg);
+        img.onload = function() {
+            ctx.drawImage(img, 0, 0);
+            resolve(ctx);
+            DOMURL.revokeObjectURL(url);
+        };
+        img.src = url;
+    })
 }
